from AutoDiff import DiffObj, Variable, Constant
from AutoDiff import MathOps as mo
import numpy as np

class NeuralNet(object):
    def __init__(self, input_dim, hidden_dim, out_dim, lr=0.1):
        self.in_var, self.in_val = self.DeclareVariables('x', (1, input_dim+1))
        self.hid_wt_var, self.hid_wt_val = self.DeclareVariables('w', (hidden_dim, input_dim+1))
        self.hid_function = self.AffineLayer(self.in_var[0].values(), self.hid_wt_var, \
                                             activation=True)
        self.hid_var, self.hid_val = self.DeclareVariables('h', (1, hidden_dim))
        self.out_wt_var, self.out_wt_val = self.DeclareVariables('v', (out_dim, hidden_dim))
        self.out_function = self.AffineLayer(self.hid_var[0].values(), self.out_wt_var)
        self.out_var, self.out_val = self.DeclareVariables('z', (1, out_dim))
        self.out_true_var, self.out_true_val = self.DeclareVariables('y', (out_dim, 1))
        self.loss = self.GetLoss('l2')
        self.lr = lr
    
    @classmethod
    def Sigmoid(cls, x):
        return 1/(1 + mo.exp(-x))
    
    def DeclareVariables(self, base_name, dim):
        var = [{base_name + '_' + str(i+1) + ('_' + str(j+1) if dim[1] > 1 else '') : \
               Variable(base_name + '_' + str(i+1) + ('_' + str(j+1) if dim[1] > 1 else '')) \
               for j in range(dim[1])} for i in range(dim[0])]
        val_dict = [{base_name + '_' + str(i+1) + ('_' + str(j+1) if dim[1] > 1 else '') : np.random.randn()\
               for j in range(dim[1])} for i in range(dim[0])]
        return var, val_dict
    
    def AffineLayer(self, in_obj, wt_var, activation=False):
        layer_out = [None]*len(wt_var)
        for idx, node_wt in enumerate(wt_var):
            layer_out[idx] = sum(a*b for a,b in zip(in_obj, node_wt.values()))
            if activation:
                layer_out[idx] = NeuralNet.Sigmoid(layer_out[idx])
        return layer_out
    
    def UpdateParams(self, wt_val, grad):
        for w, dw in zip(wt_val, grad):
            new_w = np.asarray(list(w.values())) - self.lr*np.asarray(dw)
            w.update(zip(w.keys(), new_w))
    
    def GetLoss(self, loss_fn='l2'):
        loss_obj = None
        if loss_fn == 'l2':
            loss_obj = sum((a - b)**2 for a,b in zip(self.out_true_var[0].values(), \
                                              self.out_var[0].values()))
        return loss_obj
    
    def forward(self, X, y):
        X_with_bias = np.array([1] + list(X))
        self.in_val[0].update(zip(self.in_val[0].keys(), X_with_bias))
        hid_vals = []
        for idx, hid_wt in enumerate(self.hid_wt_val):
            val_dict = {**self.in_val[0], **hid_wt}
            hid_vals += [self.hid_function[idx].get_val(val_dict)]
        self.hid_val[0].update(zip(self.hid_val[0].keys(), hid_vals))
        out_vals = []
        for idx, out_wt in enumerate(self.out_wt_val):
            val_dict = {**self.hid_val[0], **out_wt}
            out_vals += [self.out_function[idx].get_val(val_dict)]
        self.out_val[0].update(zip(self.out_val[0].keys(), out_vals))
        self.out_true_val[0].update(zip(self.out_true_val[0].keys(), [y]))
        val_dict = {**self.out_val[0], **self.out_true_val[0]}
        loss = self.loss.get_val(val_dict)
        return loss, self.out_val[0].values()
        
    def backward(self):
        val_dict = {**self.out_val[0], **self.out_true_val[0]}
        dLdz = self.loss.get_der(val_dict, with_respect_to=self.out_val[0].keys())
        dzdv, dzdh = [], []
        for idx, out_wt in enumerate(self.out_wt_val):
            val_dict = {**self.hid_val[0], **out_wt}
            dzdv += [self.out_function[idx].get_der(val_dict, \
                                                 with_respect_to=out_wt.keys())]
            dzdh += [self.out_function[idx].get_der(val_dict, \
                                                 with_respect_to=self.hid_val[0].keys())]
        dLdv =  [list(dLdz.values())[0]*np.asarray(list(dzdv[0].values()))]
        dLdh =  list(dLdz.values())[0]*np.asarray(list(dzdh[0].values()))[:, np.newaxis]
        dhdw = []
        for idx, hid_wt in enumerate(self.hid_wt_val):
            val_dict = {**self.in_val[0], **hid_wt}
            dhdw += [list(self.hid_function[idx].get_der(val_dict, \
                                                   with_respect_to=hid_wt.keys()).values())]
        dhdw = np.asarray(dhdw)
        dLdw = dLdh*dhdw
        self.UpdateParams(self.out_wt_val, dLdv)
        self.UpdateParams(self.hid_wt_val, dLdw)


import math
import numpy as np

class DiffObj(object):
    '''
    All functions will be represented by an instance of this class DiffObj, or by instances of
    classes which inherit from DiffObj (e.g. class Variable, class Constant etc.) DiffObj enforces
    that each class which inherits from it, must implement two functions:

    CLASS FUNCTIONS
    ==================
    The functions get_val and get_der are exposed to the user, that is, a user of our package can
    call these functions.

    (1) get_val:        This is used to evaluate the function represented by a DiffObj instance at
                        a particular point.
                        
    (2) get_der:        This is used to evalate the gradient of the function repreesnted by a DiffObj
                        instance, at a particular point.

    CLASS ATTRIBUTES
    ================
    The attributes are not meant to be used by an end-user of our package, and they are meant for internal
    computation.

    name_list:          A list of strings, where each item in the list represents the variables inside
                        the function represented by this DiffObj. E.g. for f(x,y) = x + y, the name_list
                        for a DiffObj representing f will be ['x', 'y'] (assuming the x.name_list = ['x']
                        and y.name_list = ['y'].
    operator:           A single string representing the "operator". By default, DiffObj assumes that it
                        represents two DiffObj's connected by an binary operator such as 'add'. However, 
                        we use the same definition for unary operators such as negation or cosine.
    operand_list:       A list of two DiffObjs, which together with self.operator, comprise this instance
                        of DiffObj.
    '''
    OVERLOADED_OPERATORS = ['add', 'subtract', 'multiply', 'divide',
            'power', 'neg', 'rdivide', 'rpower']
    def __init__(self, name_list, operator, operand_list):
        self.name_list = name_list
        self.operator = operator
        self.operand_list = operand_list
    def get_val(self, value_dict):
        '''
        INPUT
        ======
        value_dict:     A dictionary, whose keys are strings representing variables which feature
                        in the formula represented by this DiffObj. The values at those keys are
                        the values at which the formula representing this DiffObj will be evaluated.

                        E.g. For a DiffObj which represents the function f(x,y) = x + y, the value_dict
                        argument may look like value_dict = {'x': 10, 'y': 5}
        OUTPUT
        ======

        DOCTEST
        ======
        >>> z=x+y
        >>> z.get_val({'x':1,'y':1})
        2


        result:         A floating point number, which equals the evaluation of the function
                        represented by this DiffObj, at the variable values given by val_dict.
        '''
        if self.operator not in DiffObj.OVERLOADED_OPERATORS:
            raise ValueError('{} is not a supported operator'.format(self.operator))
        try:
            op1_val = self.operand_list[0].get_val(value_dict)
        except AttributeError:
            op1_val = self.operand_list[0]
        
        try:
            op2_val = self.operand_list[1].get_val(value_dict)
        except AttributeError:
            op2_val = self.operand_list[1]
        
        if self.operator == 'add':
            return op1_val + op2_val
        elif self.operator == 'subtract':
            return op1_val - op2_val
        elif self.operator == 'multiply':
            return op1_val*op2_val
        elif self.operator == 'divide':
            try:
                result = op1_val/op2_val
                return result
            except:
                raise ValueError('Division by zeros is not allowed')
        elif self.operator == 'rdivide':
            try:
                result = op2_val/op1_val
                return result
            except:
                raise ValueError('Division by zeros is not allowed')
        elif self.operator == 'power':
            return op1_val**op2_val
        elif self.operator == 'rpower':
            return op2_val**op1_val
        elif self.operator == 'neg':
            return -op1_val

    def get_der(self, value_dict, with_respect_to=None):
        '''
        INPUT
        ======
        value_dict:         A dictionary, whose keys are strings representing variables which feature
                            in the formula represented by this DiffObj. The values at those keys are
                            the values at which the gradient of formula representing this DiffObj will 
                            be evaluated.
                            
                            E.g. For a DiffObj which represents the function f(x,y) = x + y, the value_dict
                            argument may look like value_dict = {'x': 10, 'y': 5}
        OUTPUT
        ======
        result:             A dictionary, whose keys are strings representing variables which feature 
                            in the formula represented by this DiffObj. The value associated withe each
                            key is a floating point number which is the partial derivative of this DiffObj 
                            with respect to that variable.
        with_respect_to:    A list of strings representing variables, with respect to which we want the 
                            gradient of this DifObj. By default, if this list is not provided, then the
                            gradient with respect to all variables featuring in the DiffObj is returned.

        DOCTEST
        ======
        >>> z=x+y
        >>> z.get_der({'x':0,'y':0})
        {'y': 1, 'x': 1}

        '''
        if not with_respect_to: with_respect_to = self.name_list
        df = {}
        op1, op2 = self.operand_list[0], self.operand_list[1]
        try:
            op1_val = op1.get_val(value_dict)
        except AttributeError:
            op1_val = op1
        try:
            op2_val = op2.get_val(value_dict)
        except AttributeError:
            op2_val = op2

        if self.operator == 'add':
            for w in with_respect_to:
                try:
                    op2_der = op2.get_der(value_dict, [w])[w]
                except AttributeError:
                    op2_der = 0
                dw = op1.get_der(value_dict, [w])[w] + op2_der
                df[w] = dw
        elif self.operator == 'subtract':
            for w in with_respect_to:
                try:
                    op2_der = op2.get_der(value_dict, [w])[w]
                except AttributeError:
                    op2_der = 0 
                dw = op1.get_der(value_dict, [w])[w] - op2_der
                df[w] = dw
        elif self.operator == 'multiply':
            for w in with_respect_to:
                try:
                    op2_der = op2.get_der(value_dict, [w])[w]
                except AttributeError:
                    op2_der = 0  
                dw = op1.get_der(value_dict, [w])[w]*op2_val + op2_der*op1_val
                df[w] = dw
        elif self.operator == 'divide':
            try:
                one_by_op2_val = 1.0/op2_val
            except:
                raise ValueError('Division by zero is not allowed')
            for w in with_respect_to:
                try:
                    op2_der = op2.get_der(value_dict, [w])[w]
                except AttributeError:
                    op2_der = 0  
                try:
                    op1_der = op1.get_der(value_dict, [w])[w]
                except AttributeError:
                    op1_der = 0  
                dw = (op2_val*op1_der - op1_val*op2_der)/(op2_val**2)
                df[w] = dw
        elif self.operator == 'rdivide':
            try:
                one_by_op1_val = 1.0/op1_val
            except:
                raise ValueError('Division by zero is not allowed')
            for w in with_respect_to:
                try:
                    op2_der = op2.get_der(value_dict, [w])[w]
                except AttributeError:
                    op2_der = 0  
                try:
                    op1_der = op1.get_der(value_dict, [w])[w]
                except AttributeError:
                    op1_der = 0  
                dw = (op1_val*op2_der - op2_val*op1_der)/(op1_val**2)
                df[w] = dw
        elif self.operator == 'power':
            for w in with_respect_to:
                try:
                    op2_der = op2.get_der(value_dict, [w])[w]
                except AttributeError:
                    op2_der = 0  
                try:
                    op1_der = op1.get_der(value_dict, [w])[w]
                except AttributeError:
                    op1_der = 0  
                if op1_val < 0:
                    if (op2_val/1.0).is_integer() and op2_der == 0:
                        func_val = op1_val**op2_val
                        dw = func_val*op2_val/op1_val*op1_der
                    else:
                        raise ValueError('Base in Exponentiation should be positive.')
                else:
                    func_val = op1_val**op2_val
                    dw = func_val*((op2_val/op1_val)*op1_der + math.log(op1_val)*op2_der)
                df[w] = dw
        elif self.operator == 'rpower':
            for w in with_respect_to:
                try:
                    op2_der = op2.get_der(value_dict, [w])[w]
                except AttributeError:
                    op2_der = 0  
                try:
                    op1_der = op1.get_der(value_dict, [w])[w]
                except AttributeError:
                    op1_der = 0  
                if op2_val < 0:
                    if (op1_val/1.0).is_integer() and op1_der == 0:
                        func_val = op2_val**op1_val
                        dw = func_val*op1_val/op2_val*op2_der
                    else:
                        raise ValueError('Base in Exponentiation should be positive.')
                else:
                    func_val = op2_val**op1_val
                    dw = func_val*((op1_val/op2_val)*op2_der + math.log(op2_val)*op1_der)
                df[w] = dw
        elif self.operator == 'neg':
            for w in with_respect_to:
                dw = -op1.get_der(value_dict, [w])[w]
                df[w] = dw
        if len(df) == 0: df = {'' : 0}
        return df

    def getBinaryOperator(self, other, operator_name):
        try:
            this_name_list = self.name_list
        except AttributeError:
            this_name_list = []
        try:
            other_name_list = other.name_list
        except AttributeError:
            other_name_list = []
        if operator_name == 'neg': other_name_list = []
        return DiffObj(this_name_list + other_name_list,
                operator_name, [self, other])

    def __neg__(self):
        '''
        Overloads negation for objects of type DiffObj.
        INPUT
        =====
        Takes a single AutoDiff object (can be of type AutoDiff.DiffObj, AutoDiff.Constant, AutoDiff.Variable, or AutoDiff.MathOps):

        a = AutoDiff object

        -a

        which uses our __neg__ method.
        a.__neg__()
        
        OUTPUT
        ======
        result:         A DiffObj, for which DiffObj.operator_name is 'neg', DiffObj.operand_list 
                        contains [a,a], and DiffObj.name_list is same as the original name_list of a.

        DOCTEST
        ======
        >>> z=-y
        >>> z.get_val({'x':3,'y':2})
        -2
        >>> z.get_der({'x':3,'y':2})
        {'y': -1}
                        
        '''
        return self.getBinaryOperator(self, 'neg')
    def __add__(self, other):
        '''
        Overloads the add operator such that it works for DiffObj objects.

        INPUT
        =====
        Takes two AutoDiff objects (can be of type AutoDiff.DiffObj, AutoDiff.Constant, AutoDiff.Variable, or AutoDiff.MathOps):
    
        a = AutoDiff object
        b = AutoDiff object

        a + b

        which uses our __add__ method:

        a.__add__(b)

        OUTPUT
        ======
        Returns a DiffObj, where the DiffObj.name_list is the concatination of a.name_list and b.name_list,
        DiffObj.operator_name is 'add', and DiffObj.operand_list is [a,b].

        When get_val is called on the resulting DiffObj, as in DiffObj.get_val(val_dict), the sum
        DiffObj.operand_list[0].get_val(val_dict) + DiffObj.operand_list[1].get_val(val_dict) is returned.
        In other words, DiffObj.get_val(val_dict) returns the sum of the operands after
        their values are evaluated individually with respect to their own value dictionaries.

        DOCTEST
        ======
        >>> z=x+y
        >>> z.get_val({'x':3,'y':2})
        5
        >>> z.get_der({'x':3,'y':2})
        {'y': 1, 'x': 1}

        '''


        return self.getBinaryOperator(other, 'add')
    def __sub__(self, other):
        '''
        Overloads the subtract operator such that it works for DiffObj objects.

        INPUT
        =====
        Takes two AutoDiff objects (can be of type AutoDiff.DiffObj, AutoDiff.Constant, AutoDiff.Variable, or AutoDiff.MathOps):

        a = AutoDiff object
        b = AutoDiff object

        a - b

        which uses our __sub__ method:

        a.__sub__(b)

        OUTPUT
        ======
        Returns a DiffObj, where the DiffObj.name_list is the concatination of a.name_list and b.name_list,
        DiffObj.operator_name is 'subtract', and DiffObj.operand_list is [a,b].

        When get_val is called on the resulting DiffObj, as in DiffObj.get_val(val_dict), the difference
        DiffObj.operand_list[0].get_val(val_dict) - DiffObj.operand_list[1].get_val(val_dict) is returned.
        In other words, DiffObj.get_val(val_dict) returns the difference of the operands after
        their values are evaluated individually with respect to their own value dictionaries.

        DOCTEST
        ======
        >>> z=x-y
        >>> z.get_val({'x':3,'y':2})
        1
        >>> z.get_der({'x':3,'y':2})
        {'y': -1, 'x': 1}

        '''
        return self.getBinaryOperator(other, 'subtract')

    def __rsub__(self, other):
        return -self + other
    
    def __mul__(self, other):
        '''
        Overloads the multiply operator such that it works for DiffObj objects.

        INPUT
        =====
        Takes two AutoDiff objects (can be of type AutoDiff.DiffObj, AutoDiff.Constant, AutoDiff.Variable, or AutoDiff.MathOps):

        a = AutoDiff object
        b = AutoDiff object

        a * b

        which uses our __mul__ method:

        a.__mul__(b)

        OUTPUT
        ======  
        Returns a DiffObj, where the DiffObj.name_list is the concatination of a.name_list and b.name_list,
        DiffObj.operator_name is 'multiply', and DiffObj.operand_list is [a,b].

        When get_val is called on the resulting DiffObj, as in DiffObj.get_val(val_dict), the product
        DiffObj.operand_list[0].get_val(val_dict) * DiffObj.operand_list[1].get_val(val_dict) is returned.
        In other words, DiffObj.get_val(val_dict) returns the product of the operands after
        their values are evaluated individually with respect to their own value dictionaries.


        DOCTEST
        ======
        >>> z=x*y
        >>> z.get_val({'x':3,'y':2})
        6
        >>> z.get_der({'x':3,'y':2})
        {'y': 3, 'x': 2}

        '''
        return self.getBinaryOperator(other, 'multiply')

    def __truediv__(self, other):
        '''
        Overloads the division operator such that it works for DiffObj objects.

        INPUT
        =====
        Takes two AutoDiff objects (can be of type AutoDiff.DiffObj, AutoDiff.Constant, AutoDiff.Variable, or AutoDiff.MathOps):

        a = AutoDiff object
        b = AutoDiff object

        a / b

        which uses our __truediv__ method:

        a.__truediv__(b)

        OUTPUT
        ======
        Returns a DiffObj, where the DiffObj.name_list is the concatination of a.name_list and b.name_list,
        DiffObj.operator_name is 'divide', and DiffObj.operand_list is [a,b].

        When get_val is called on the resulting DiffObj, as in DiffObj.get_val(val_dict), the division
        DiffObj.operand_list[0].get_val(val_dict) / DiffObj.operand_list[1].get_val(val_dict) is returned.
        In other words, DiffObj.get_val(val_dict) returns the division of the operands after
        their values are evaluated individually with respect to their own value dictionaries.

        DOCTEST
        ======
        >>> z=x/y
        >>> z.get_val({'x':3,'y':2})
        1
        >>> z.get_der({'x':3,'y':2})
        {'y': -1, 'x': 0}
        >>> 


        '''
        return self.getBinaryOperator(other, 'divide')

    
    def __rtruediv__(self, other): 
        return self.getBinaryOperator(other, 'rdivide')

    def __div__(self,other):
        return self.__truediv__(other)
        '''
        __div__ for python2 support


        DOCTEST
        ======
        >>> z=x/y
        >>> z.get_val({'x':3,'y':2})
        1
        >>> z.get_der({'x':3,'y':2})
        {'y': -1, 'x': 0}
        >>> 

        '''

    def __pow__(self, other):
        '''
        Overloads the power operator such that it works for DiffObj objects.
        INPUT
        =====
        self, other:        Two AutoDiff objects (can be of type AutoDiff.DiffObj, AutoDiff.Constant, 
                            AutoDiff.Variable, or AutoDiff.MathOps

        Example Usage:
        If a and b are two AutoDiff Objects. Then a**b will use our __pow__ method.


        OUTPUT
        ======
        result:             A DiffObj where DiffObj.name_list is the concatenation of a.name_list and 
                            b.name_list, DiffObj.operator_name is 'power', and DiffObj.operand_list is
                            [a,b].

        DOCTEST
        ======
        >>> z=x*y
        >>> z.get_val({'x':3,'y':2})
        6
        >>> z.get_der({'x':3,'y':2})
        {'y': 3, 'x': 2}
        '''
        return self.getBinaryOperator(other, 'power')

    def __rpow__(self, other):
        return self.getBinaryOperator(other, 'rpower')

    __radd__ = __add__
    __rmul__ = __mul__
    
   
class Variable(DiffObj):
    '''
    All variables inside a function whose derivative and value a user wants to calculate, 
    will be instances of the Variable class, which inherits from DiffObj and implements
    get_val and get_der

    CLASS ATTRIBUTES
    ================
    var_name:           A string, which is unique to this Variable instance.
                        E.g. x = Variable('x')

    CLASS FUNCTIONS
    ===============
    This implements get_val and get_der, a description of which is provided in the 
    Super-class DiffObj.

    '''
    def __init__(self, var_name):
        self.var_name = var_name
        super(Variable, self).__init__([var_name], None, None)
    
    def get_val(self, value_dict):
        if self.var_name not in value_dict:
            raise ValueError('You have not provided a value for {}'.format(self.var_name))
        try:
            temp = value_dict[self.var_name] + 0.0
        except:
            raise TypeError('Only integer and float types are accepted as values.')
        return value_dict[self.var_name]
    
    def get_der(self, value_dict, with_respect_to=None):
        if not with_respect_to:
            return {self.var_name : 1}
        else:
            der_dict = {}
            for w in with_respect_to:
                der_dict[w] = int(w == self.var_name)
            return der_dict

class Constant(DiffObj):
    '''
    All constants inside a function whose derivative and value a user wants to calculate,
    will be instances of the Constant class, which inherits from DiffObj and implements
    get_val and get_der

    CLASS ATTRIBUTES
    ================
    const_name:         A string, which is unique to this Constant instance.
    const_val:          An int or float number, which will be the value assigned to this instance.

                        E.g. c = Constant('c', 10.0)

    CLASS FUNCTIONS
    ===============
    This implements get_val and get_der, a description of which is provided in the
    Super-class DiffObj. As expected, get_val simply returns self.const_val while
    get_der will return 0.
    '''
    def __init__(self, const_name, const_val):
        super(Constant, self).__init__([], None, None)
        self.const_name = const_name
        self.const_val = const_val
        self.name_list = []
    def get_val(self, value_dict):
        return self.const_val
    def get_der(self, value_dict, with_respect_to=None):
        if not with_respect_to:
            return {'' : 0}
        der_dict = {}
        for w in with_respect_to:
            der_dict[w] = 0
        return der_dict


class VectorFunction(DiffObj):
    '''
    Temporary vector function for use in root_finder.py until code supports vector valued functions
    ''' 
    def __init__(self, list_of_functions):
        self.list_of_functions = list_of_functions
        name_list = []
        for f in self.list_of_functions:
            name_list = name_list+f.name_list
        name_list = list(set(name_list))
        super(VectorFunction, self).__init__(name_list, None, None)


    def get_val(self, val_dict):
        return [f.get_val(val_dict) for f in self.list_of_functions]  
    def get_der(self, val_dict, with_respect_to=None):
        return [f.get_der(val_dict, with_respect_to) for f in self.list_of_functions] 

    # returns 2d array from list of dictionarys
    def dict_list_to_array(self,dict_list):
        arr = []
        for f_dict in dict_list:
            row = []
            for var in self.name_list:
                row.append(f_dict[var])
            arr.append(row)        
        return arr



class MathOps(DiffObj):
    '''
    This class inherits from the DiffObj class. It implements non-elementary unary functions 
    including: sin, cos, tan, log, exp.

    INSTANTIATION
    ===============
    If a is of type DiffObj, then the invoking the constructor as follows will return an 
    object b of type MathOps:

    b = MathOps.sin(a)

    CLASS ATTRIBUTES
    ================
    The attributes are not meant to be used by an end-user of our package, and they are meant for internal
    computation.

    name_list:          A list of strings, where each item in the list represents the variables inside
                        the function represented by this DiffObj. E.g. for f(x,y) = x + y, the name_list
                        for a DiffObj representing f will be ['x', 'y'] (assuming the x.name_list = ['x']
                        and y.name_list = ['y'].
    operator:           A string, such as 'sin' or 'log', which represents one of the unary math operators
                        implemented by this class.
    operand_list:       A list of length 1 containing the DiffObj which the user has passed as an argument
                        to one of the classmethods of MathOps.

    '''
    def __init__(self, name_list, operator, operand):
        super(MathOps, self).__init__(name_list, 
                operator, operand)
    @classmethod
    def getUnaryOperator(cls, operator, obj):
        try:
            name_list = obj.name_list
            return MathOps(name_list, operator, [obj]) 
        except:
            raise TypeError('Only objects of type DiffObj are permitted.')
    @classmethod
    def sin(cls, obj):
        '''
        INPUT
        =====
        obj:        An object of type DiffObj, on which the user wants to
                    apply the sine function.
        OUTPUT
        ======
        result:     A DiffObj, whose operator is 'sin' and whose operand is
                    the DiffObj on which the user had called this sin function.

        DOCTEST
        ======
        >>> z=MathOps.sin(x)
        >>> z.get_val({'x':math.pi})
        1.2246467991473532e-16
        >>> z.get_der({'x':math.pi})
        {'x': -1.0}
        
        '''
        return MathOps.getUnaryOperator('sin', obj)
    @classmethod
    def cos(cls, obj):
        '''
        INPUT
        =====
        obj:        An object of type DiffObj, on which the user wants to
                    apply the cos function.
        OUTPUT
        ======
        result:     A DiffObj, whose operator is 'cos' and whose operand is
                    the DiffObj on which the user had called this cos function.
        DOCTEST
        ======
        
        >>> z=MathOps.cos(x)
        >>> z.get_val({'x':math.pi})
        -1.0
        >>> z.get_der({'x':math.pi})
        {'x': -1.2246467991473532e-16}
        '''
        return MathOps.getUnaryOperator('cos', obj)
    @classmethod
    def tan(cls,obj):
        '''
        INPUT
        =====
        obj:        An object of type DiffObj, on which the user wants to
                    apply the tan function.
        OUTPUT
        ======
        result:     A DiffObj, whose operator is 'sin' and whose operand is
                    the DiffObj on which the user had called this tan function.

        DOCTEST
        ======
        
        >>> z=MathOps.tan(x)
        >>> z.get_val({'x':0})
        0.0
        >>> z.get_der({'x':0})
        {'x': 1.0}
        '''

        return MathOps.getUnaryOperator('tan', obj)
    @classmethod
    def log(cls, obj):
        '''
        INPUT
        =====
        obj:        An object of type DiffObj, on which the user wants to
                    apply the natural log function.
        OUTPUT
        ======
        result:     A DiffObj, whose operator is 'sin' and whose operand is
                    the DiffObj on which the user had called this log function.

        DOCTEST
        ======
        >>> z=MathOps.log(x)
        >>> z.get_val({'x':1})
        0.0
        >>> z.get_der({'x':1})
        {'x': 1.0}
        '''

        return MathOps.getUnaryOperator('log', obj)
    @classmethod
    def exp(cls, obj):
        '''
        INPUT
        =====
        obj:        An object of type DiffObj, on which the user wants to
                    apply the natural exponentiation function.
        OUTPUT
        ======
        result:     A DiffObj, whose operator is 'sin' and whose operand is
                    the DiffObj on which the user had called this exp function.
        >>> z=MathOps.exp(x)
        >>> z.get_val({'x':0})
        1.0
        >>> z.get_der({'x':0})
        {'x': 1.0}
        '''

        return MathOps.getUnaryOperator('exp', obj)
    def get_val(self, value_dict):
        '''
        INPUT
        ======
        value_dict:     A dictionary, whose keys are strings representing variables which feature
                        in the formula represented by this DiffObj. The values at those keys are
                        the values at which the formula representing this DiffObj will be evaluated.

                        E.g. For a DiffObj which represents the function f(x,y) = x + y, the value_dict
                        argument may look like value_dict = {'x': 10, 'y': 5}
        OUTPUT
        ======
        result:         A floating point number, which equals the evaluation of the function
                        represented by this DiffObj, at the variable values given by val_dict.
        '''
 
        operand_val = self.operand_list[0].get_val(value_dict)
        if self.operator == 'sin':
            return math.sin(operand_val)
        elif self.operator == 'cos':
            return math.cos(operand_val)
        elif self.operator == 'tan':
            result = math.tan(operand_val)
            return result
        elif self.operator == 'log':
            try:
                result = math.log(operand_val)
                return result
            except:
                raise ValueError('Only positive values are permitted with log.')
        elif self.operator == 'exp':
            result = math.exp(operand_val)
            return result

    def get_der(self, value_dict, with_respect_to=None):
        '''
        INPUT
        ======
        value_dict:         A dictionary, whose keys are strings representing variables which feature
                            in the formula represented by this DiffObj. The values at those keys are
                            the values at which the gradient of formula representing this DiffObj will 
                            be evaluated.
                            
                            E.g. For a DiffObj which represents the function f(x,y) = x + y, the value_dict
                            argument may look like value_dict = {'x': 10, 'y': 5}
        OUTPUT
        ======
        result:             A dictionary, whose keys are strings representing variables which feature 
                            in the formula represented by this DiffObj. The value associated withe each
                            key is a floating point number which is the partial derivative of this DiffObj 
                            with respect to that variable.
        with_respect_to:    A list of strings representing variables, with respect to which we want the 
                            gradient of this DifObj. By default, if this list is not provided, then the
                            gradient with respect to all variables featuring in the DiffObj is returned.
        '''
 
        if not with_respect_to: with_respect_to = self.name_list
        df = {}
        op1 = self.operand_list[0]
        if self.operator == 'sin':
            for w in with_respect_to:
                dw = math.cos(op1.get_val(value_dict))*op1.get_der(value_dict, [w])[w]
                df[w] = dw
        elif self.operator == 'cos':
            for w in with_respect_to:
                dw = -math.sin(op1.get_val(value_dict))*op1.get_der(value_dict, [w])[w]
                df[w] = dw
        elif self.operator == 'tan':
            sec_x = 1.0/math.cos(op1.get_val(value_dict))
            for w in with_respect_to:
                dw = (sec_x**2)*op1.get_der(value_dict, [w])[w]
                df[w] = dw
        elif self.operator == 'log':
            try:
                one_by_var = 1.0/op1.get_val(value_dict)
            except:
                raise ValueError('Log cannot be evaluated at 0.')
            for w in with_respect_to:
                dw = one_by_var*op1.get_der(value_dict, [w])[w]
                df[w] = dw
        elif self.operator == 'exp':
            func_val = math.exp(op1.get_val(value_dict))
            for w in with_respect_to:
                dw = func_val*op1.get_der(value_dict, [w])[w]
                df[w] = dw

        if len(df) == 0: df = {'' : 0}
        return df

if __name__ == "__main__":
    import doctest
    doctest.testmod(extraglobs={'x': Variable('x'),'y':Variable('y')})



from AutoDiff import api
from threading import Thread
import random
import numpy as np
import math
import logging


# create a thread object that returns the thread results
# adapted from kindall on stackoverflow
class ThreadWithReturnValue(Thread):
    def __init__(self, group=None, target=None, name=None,
                 args=()):
        Thread.__init__(self, group, target, name, args)
        self._return = None
    def run(self):
        if self._target is not None:
            self._return = self._target(*self._args)
    def join(self, *args):
        Thread.join(self, *args)
        return self._return

# multi-dimensional Newton's method
def vectorNewton(input_function,tolerance=1e-5, num_starting_vals = 20, 
	starting_val_range = (-1000,1000), starting_val_dict_list=None, verbose=True):

	logger = logging.getLogger()

	# initialize our list
	if not starting_val_dict_list:
		starting_val_dict_list = []

	# find one root 
	def find_root(vf,starting_val_dict, max_iter,tol):
		
		val_dict = starting_val_dict

		error_list = []
		fx = f.get_val(val_dict)
		it = 0
		all_dim_dist = math.inf

		# if we haven't done too many iterations and we're still greater than our tolerance 
		while it < max_iter and all_dim_dist > tol:
			it += 1

			abs_fx = [abs(val) for val in fx]
			all_dim_dist = np.sum(abs_fx)
			all_dim_dist = np.linalg.norm(fx,ord=2)

			# make a list of errors so we can check Newton's method is working
			error_list += [np.sum(abs_fx)]

			# get jacobian, move to new point
			try:
				dx = vf.dict_list_to_array(vf.get_der(val_dict))

				# need to update all new values of value dictionary
				delta = np.linalg.solve(np.array(dx),-1*np.array(fx))

				# update dictionary
				for i,val in enumerate(vf.name_list):
					val_dict[val] = val_dict[val] + delta[i]
				new_fx = vf.get_val(val_dict)
				fx = new_fx

			# avoid dividing by zero 
			except:
				logger.warning("Tried to divide by zero!")
				return

		return ([val_dict[var] for var in vf.name_list], vf.get_val(val_dict), i, error_list)

	# function takes value and list, returns true if value is within diff_tol of any value
	# in the list, false otherwise.
	def is_close_vector_lists(v,lst,diff_tol=1e-6):
		for ele in lst:
			l = [abs(v[i]-ele[i]) for i in range(len(v))]
			if np.sum(l)<diff_tol:
				return True
		return False


	f = input_function

	# adjust starting value list to agree with number of requested starting values 
	while len(starting_val_dict_list)<num_starting_vals:
		starting_val_dict_to_add = {}
		for var in f.name_list:
			starting_val_dict_to_add[var] = random.randint(starting_val_range[0],starting_val_range[1])
		starting_val_dict_list.append(starting_val_dict_to_add)

	max_iter = 100 # maybe user should be able to choose this? 
	results = []
	roots = []

	# start threads 
	threads = []
	for i in range(len(starting_val_dict_list)):
		thread = ThreadWithReturnValue(target=find_root, 
			args=(f,starting_val_dict_list[i],max_iter,tolerance))
		thread.start()
		threads.append(thread)
	for i in range(len(starting_val_dict_list)):
		full_result = threads[i].join()

		# if didn't catch exception in find_root 
		if full_result:
			root_result = full_result[0]

			is_zero = True
			for output in full_result[1]:
				if abs(output) > tolerance:
					is_zero = False

			# check if root already in list 
			if not (is_close_vector_lists(root_result, roots)) and is_zero:
				roots.append(root_result)
				results.append(full_result)

	# for testing, choose verbose 
	if verbose:
		return results
	else:
		return roots


def quasi_newton(input_function,tolerance):
	pass


def multiple_function_root_finder(input_function, lr=0.01, adadelta=False):
	pass


import pytest 
import math
from AutoDiff import DiffObj, Variable, Constant
from AutoDiff import MathOps as mo

class TestAutoDiff():

	# test add 
	def test_add(self):  
		val_dict = {'x' : 10, 'y' : 20, 'z':1}
		x = Variable('x')
		y = Variable('y')
		z = Variable('z')
		c1 = Constant('c1', 5)

		f0 = x + y		     
		assert(f0.get_val(val_dict) ==30)
		assert(f0.get_der(val_dict)['x'] == 1)
		assert(f0.get_der(val_dict)['y'] == 1)

		f1 = y + x
		assert(f1.get_val(val_dict) ==30)
		assert(f1.get_der(val_dict)['x'] == 1)
		assert(f1.get_der(val_dict)['y'] == 1)	

		f2 = f0 + f1 + x
		assert(f2.get_val(val_dict) ==70)
		assert(f2.get_der(val_dict)['x'] == 3)
		assert(f2.get_der(val_dict)['y'] == 2)

		f3 = f0 + c1	
		assert(f3.get_val(val_dict) ==35)
		assert(f3.get_der(val_dict)['x'] == 1)
		assert(f3.get_der(val_dict)['y'] == 1)

		f4 = c1 + f0	
		assert(f4.get_val(val_dict) ==35)
		assert(f4.get_der(val_dict)['x'] == 1)
		assert(f4.get_der(val_dict)['y'] == 1)

		assert(c1.get_der(val_dict, ['x', 'y'])['x'] == 0)
		assert(c1.get_der(val_dict, ['x', 'y'])['y'] == 0)


	# test subtract 
	def test_subtract(self):
		val_dict = {'x' : 10, 'y' : 20}
		x = Variable('x')
		y = Variable('y')
		f0 = x - y	
		c1 = Constant('c1', 5)

		assert(f0.get_val(val_dict) ==-10)
		assert(f0.get_der(val_dict)['x'] == 1)
		assert(f0.get_der(val_dict)['y'] == -1)

		f1 = y - x
		assert(f1.get_val(val_dict) ==10)
		assert(f1.get_der(val_dict)['x'] == -1)
		assert(f1.get_der(val_dict)['y'] == 1)

		f2 = f0 + f1 - x
		assert(f2.get_val(val_dict) ==-10)
		assert(f2.get_der(val_dict)['x'] == -1)
		assert(f2.get_der(val_dict)['y'] == 0)

		f3 = f0 - c1
		assert(f3.get_val(val_dict) == -15)
		assert(f3.get_der(val_dict)['x'] == 1)
		assert(f3.get_der(val_dict)['y'] == -1)

		f4 = c1 - f0
		assert(f4.get_val(val_dict) == 15)
		assert(f4.get_der(val_dict)['x'] == -1)
		assert(f4.get_der(val_dict)['y'] == 1)

		f5 = x + y - y 
		assert(f5.get_val(val_dict) ==10)
		assert(f5.get_der(val_dict)['x'] == 1)
		assert(f5.get_der(val_dict)['y'] == 0)


	# test multiply
	def test_multiply(self):
		val_dict = {'x' : 10, 'y' : 20}
		x = Variable('x')
		y = Variable('y')	
		c1 = Constant('c1', 5)

		f0 = x*y	     
		assert(f0.get_val(val_dict) ==200)
		assert(f0.get_der(val_dict)['x'] == 20)
		assert(f0.get_der(val_dict)['y'] == 10)

		f1 = y*x	     
		assert(f1.get_val(val_dict) ==200)
		assert(f1.get_der(val_dict)['x'] == 20)
		assert(f1.get_der(val_dict)['y'] == 10)

		f2 = f0*f1
		assert(f2.get_val(val_dict) ==40000)
		assert(f2.get_der(val_dict)['x'] == 8000)
		assert(f2.get_der(val_dict)['y'] == 4000)

		f3 = f0*c1 
		assert(f3.get_val(val_dict) ==1000)
		assert(f3.get_der(val_dict)['x'] == 100)
		assert(f3.get_der(val_dict)['y'] == 50)

		f4 = c1*f0 
		assert(f4.get_val(val_dict) ==1000)
		assert(f4.get_der(val_dict)['x'] == 100)
		assert(f4.get_der(val_dict)['y'] == 50)


	# test divide 
	def test_divide(self):
		val_dict = {'x' : 10, 'y' : 20}
		x = Variable('x')
		y = Variable('y')
		c1 = Constant('c1', 5)

		f0 = x/y
		assert(f0.get_val(val_dict) ==0.5)
		assert(f0.get_der(val_dict)['x'] == 0.05)
		assert(f0.get_der(val_dict)['y'] == -0.025)

		f1 = y/x
		assert(f1.get_val(val_dict) ==2.0)
		assert(f1.get_der(val_dict)['x'] == -0.2)
		assert(f1.get_der(val_dict)['y'] == 0.1)

		f2 = f1/y
		assert(f2.get_val(val_dict) ==0.1)
		assert(f2.get_der(val_dict)['x'] == -0.01)
		assert(f2.get_der(val_dict)['y'] == 0.00)

		f3 = f0/f1 
		assert(f3.get_val(val_dict) ==0.25)
		assert(f3.get_der(val_dict)['x'] == 0.05)
		assert(f3.get_der(val_dict)['y'] == -0.025)

		f4 = f0/c1 
		assert(f4.get_val(val_dict) ==0.1)
		assert(f4.get_der(val_dict)['x'] == 0.01)
		assert(f4.get_der(val_dict)['y'] == -0.005)

	# test power
	def test_power(self):
		val_dict = {'x' : 10.0, 'y' : 3.0}
		x = Variable('x')
		y = Variable('y')	
		c1 = Constant('c1', 2.0)
		c2 = Constant('c2', -2.0)
		c3 = Constant('c3', 0.0)

		f0 = x**c1
		assert(f0.get_val(val_dict) ==100)
		assert(f0.get_der(val_dict)['x'] == 20)

		f1 = x**c2
		assert(f1.get_val(val_dict) ==0.01)
		assert(f1.get_der(val_dict)['x'] == -0.002)

		f2 = x**c3 
		assert(f2.get_val(val_dict) ==1)
		assert(f2.get_der(val_dict)['x'] == 0)

		f3 = x**y
		assert(f3.get_val(val_dict) ==1000)
		assert(f3.get_der(val_dict)['x'] == 300)
		assert(f3.get_der(val_dict)['y'] == math.log(10.0)*1000)

		f4 = f0 ** f1
		assert(f2.get_val(val_dict) ==1)
		assert(f2.get_der(val_dict)['x'] == 0)


	def test_trig(self):
		val_dict = {'x' : 0, 'y' : math.pi/2}
		x = Variable('x')
		y = Variable('y')	
		c1 = Constant('c1', 2.0)
		c2 = Constant('c2', -2.0)
		c3 = Constant('c3', 0.0)

		f0 = mo.sin(x)
		assert(f0.get_val(val_dict) ==0)
		assert(f0.get_der(val_dict)['x'] == 1.0)

		f1 = c1*f0
		assert(f1.get_val(val_dict) ==0)
		assert(f1.get_der(val_dict)['x'] == 2.0)

		f2 = c1+f0
		assert(f2.get_val(val_dict) ==2.0)
		assert(f2.get_der(val_dict)['x'] == 1.0)

		f3 = mo.sin(y)
		assert(f3.get_val(val_dict) ==1.0)
		assert(f3.get_der(val_dict)['y'] == math.cos(math.pi/2))

		f4 = mo.cos(f3)
		assert(f4.get_val(val_dict) ==math.cos(math.sin(math.pi/2)))
		assert(f4.get_der(val_dict)['y'] == -math.sin(math.sin(math.pi/2))*math.cos(math.pi/2))

		f5 = mo.tan(x)
		assert(f5.get_val(val_dict)) == math.tan(0)
		assert(f5.get_der(val_dict)['x']) == 1/math.cos(0)**2



	def test_log(self):
		val_dict = {'x' : 10, 'y' : 5, 'z' : -1, 'a' : 0}
		x = Variable('x')
		y = Variable('y')
		z = Variable('z')
		a = Variable('a')
		c1 = Constant('c1',2.0)	

		f0 = mo.log(x)
		assert(f0.get_val(val_dict) ==math.log(10))
		assert(f0.get_der(val_dict)['x'] == 0.10)

		f1 = f0**c1
		assert(f1.get_val(val_dict) ==math.log(10)**2.0)
		assert(f1.get_der(val_dict)['x'] == 2*math.log(10)*0.1)

		f2 = mo.log(y)*f0
		assert(f2.get_val(val_dict) ==math.log(5)*math.log(10))
		assert(f2.get_der(val_dict)['x'] == math.log(5)*(1/10))
		assert(f2.get_der(val_dict)['y'] == math.log(10)*(1/5))

		with pytest.raises(ValueError):
			f2 = mo.log(z)
			f2.get_val(val_dict)['z']

		with pytest.raises(ValueError):
			f2 = mo.log(a)
			f2.get_der(val_dict)['a']

	def test_exp(self):
		val_dict = {'x' : 10, 'y' : 5}
		x = Variable('x')
		y = Variable('y')
		c1 = Constant('c1',2.0)	
		c2 = Constant('c2', 0.0)

		f0 = mo.exp(x)
		assert(f0.get_val(val_dict) == math.exp(10))
		assert(f0.get_der(val_dict)['x'] == math.exp(10))

		f1 = c1*f0

		assert(f1.get_val(val_dict)==2.0*math.exp(10))
		assert(f1.get_der(val_dict)['x'] == 2.0*math.exp(10))


	def test_exceptions(self):
		val_dict = {'x' : 10, 'y' : 5}
		x = Variable('x')
		y = Variable('y')
		c1 = Constant('c1',2.0)	
		c2 = Constant('c2', 0.0)
			
		with pytest.raises(ValueError):
			f0 = x/c2
			f0.get_val(val_dict)
		with pytest.raises(ValueError):
			f0.get_der(val_dict)['p']
		with pytest.raises(ValueError):
			f0.get_val({'p':5})
		with pytest.raises(TypeError):
			f0.get_val({'x':'hello'})


	def test_neg(self):
		val_dict = {'x' : 10, 'y' : 5}
		x = Variable('x')
		y = Variable('y')
		c1 = Constant('c1',2.0)	

		f1 = c1*x
		f2 = -f1
		assert(f2.get_val(val_dict) == -20.0)

		f3 = x*y
		f4 = -f3 
		assert(f4.get_val(val_dict) == -50.0)
		assert(f4.get_der(val_dict)['x']==-5)
		assert(f4.get_der(val_dict)['y']==-10)

	def test_divide_by_zero_dir(self):
		val_dict = {'x' : 0}
		x = Variable('x')
		y = Variable('y')
		c1 = Constant('c1',1/3)

		f0 = x**c1
		with pytest.raises(ZeroDivisionError):
			f0.get_der(val_dict)['x']









import pytest 
import math
import numpy as np
from AutoDiff import DiffObj, Variable, Constant, VectorFunction
from AutoDiff import MathOps as mo
from AutoDiff.root_finder import ThreadWithReturnValue, vectorNewton

TOL = 1e-6

class TestRootFinder():
    def test_no_roots(self):        
        c5 = Constant('c5', 5)
        c5_v = VectorFunction([c5])

        x = Variable('x')
        sin_x = mo.sin(x)
        sin_x_c5 = sin_x + c5
        sin_x_c5_v = VectorFunction([sin_x_c5])

        x_sq = x ** 2
        x_sq_c5 = x_sq + c5
        x_sq_c5_v = VectorFunction([x_sq_c5])

        vector_list = [c5_v, sin_x_c5_v, x_sq_c5_v]

        for i in range(len(vector_list)):
            vector = vector_list[i]
            print(i)
            roots = vectorNewton(vector, verbose=False)
            assert len(roots) == 0

    def test_single_root(self):
        c3 = Constant('c3', 3)
        x = Variable('x')

        x_cubed = x ** c3
        x_cubed_roots = vectorNewton(VectorFunction([x_cubed]), verbose=False, tolerance=TOL**4)
        print(x_cubed_roots)
        assert len(x_cubed_roots) == 1
        assert len(x_cubed_roots[0]) == 1
        assert abs(x_cubed_roots[0][0] ** 3) < TOL

        x3 = c3 * x
        x3_roots = vectorNewton(VectorFunction([x3]), verbose=False)
        assert len(x3_roots) == 1
        assert len(x3_roots[0]) == 1
        assert abs(3 * x3_roots[0][0]) < TOL

        x9 = c3 * c3 * x
        x9_roots = vectorNewton(VectorFunction([x9]), verbose=False)
        assert len(x9_roots) == 1
        assert len(x9_roots[0]) == 1
        assert abs(9 * x9_roots[0][0]) < TOL

    def test_multiple_roots(self):
        x = Variable('x')
        c1 = Constant('c1', 1)
        c2 = Constant('c2', 2)
        c3 = Constant('c3', 5)
        c4 = Constant('c4', 10)
        f = c3**(c1 + mo.sin(mo.log(c3 + x**c2))) - c4
        roots = vectorNewton(VectorFunction([f]), verbose=False)
        for root in roots:
            assert abs(5**(1+ np.sin(np.log(5 + root[0]**2))) - 10) < TOL




from setuptools import setup, find_packages

setup(
   name='AutoDiff_207_15',
   version='0.1',
   description='Autodifferentiation package',
   author='CS207 group 15',
   packages=find_packages(),  #same as name
   install_requires=['numpy'], #external packages as dependencies
   data_files = [("", ["LICENSE"])],
)